#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/serializers.py

    All known serializers are defined here, in the SERIALIZERS dict. Each serializer
    has its own serializer_xxx function.
    A reference to each module imported is added in the MODULES dict.

    ___________________________________________________________________________

    o  _len(obj)
    o  serializer_iaswn(action="serialize", obj=None, fingerprint="")
    o  serializer_json(action="serialize", obj=None, fingerprint="")
    o  serializer_jsonpickle(action="serialize", obj=None, fingerprint="")
    o  serializer_jsonpickle_keystrue(action="serialize", obj=None, fingerprint="")
    o  serializer_marshal(action="serialize", obj=None, fingerprint="")
    o  serializer_pickle(action="serialize", obj=None, fingerprint="")
    o  serializer_pyyaml(action="serialize", obj=None, fingerprint="")

    o  init_serializers()
"""
# pylint: disable = wrong-import-position

import sys
import timeit

import wisteria.globs

if wisteria.globs.PLATFORM_SYSTEM == "Windows":
    import os  # pylint: disable=unused-import
    from wmi import WMI  # pylint: disable=import-error
else:
    import resource  # pylint: disable=import-error

# MEMOVERUSE# --memoveruse C++ module:
# MEMOVERUSEimport cppyy

from wisteria.globs import TIMEITNUMBER, MODULES
from wisteria.globs import VERBOSITY_DEBUG
from wisteria.wisteriaerror import WisteriaError
from wisteria.utils import trytoimport
from wisteria.serializers_classes import SerializerData, SerializationResult
from wisteria.msg import msgdebug

# MEMOVERUSE# --memoveruse C++ module:
# MEMOVERUSEcppyy.include("memoveruse_cpp/memoveruse_cpp.h")
# MEMOVERUSEcppyy.load_library("memoveruse_cpp/libmemoveruse_cpp")
# MEMOVERUSE#    pylint: disable=import-error,wrong-import-position,wrong-import-order
# MEMOVERUSE#    about noqa: https://pep8.readthedocs.io/en/latest/intro.html#configuration
# MEMOVERUSEfrom cppyy.gbl import MemOverUse  # noqa


def _len(obj):
    """
        _len()

        Return the length of <obj>

        _______________________________________________________________________

        ARGUMENT: (bytes|str)<obj>

        RETURNED VALUE: the length of <obj>
    """
    if isinstance(obj, str):
        return len(bytes(obj, "utf-8"))
    return len(obj)


def win_memory():
    """
        win_memory()

        Compute using WMI the quantity of memory used by the current process.

        _______________________________________________________________________

        RETURNED VALUE: (int)the number of bytes used by the current process.
    """
    wmi = WMI('.')
    result = wmi.query(
        "SELECT WorkingSet "
        f"FROM Win32_PerfRawData_PerfProc_Process WHERE IDProcess={os.getpid()}")
    res = int(result[0].WorkingSet)
    return res


def serializer_iaswn(action="serialize",
                     obj=None,
                     fingerprint=""):
    """
        serializer_iaswn()

        Serializer for the Iaswn module.

        Like every serializer_xxx() function:
        * this function may return (action='version') the version of the concerned module.
        * this function may try (action='serialize') to encode/decode an <obj>ect.
        * if the serializer raises an error, this error is silently converted and no exception
          is raised. If an internal error happpens, a WisteriaError exception is raised.

        This function assumes that the concerned module has already be imported.

        _______________________________________________________________________

        ARGUMENTS:
        o  action:      (str) either "version" either "serialize"
        o  obj:         the object to be serialized
        o  fingerprint: a string describing the operation (usefull to debug)

        RETURNED VALUE:
           - None if an error occcured
           - if <action> is (str)"version", return a string.
           - if <action> is (str)"serialize", return a SerializationResult object.
    """
    module = MODULES["iaswn"]

    # -------------------
    # action == "version"
    # -------------------
    if action == "version":
        return module.__version__

    # ---------------------
    # action == "serialize"
    # ---------------------
    if action != "serialize":
        raise WisteriaError(f"(ERRORID030) Unknown 'action' keyword '{action}'.")

    # MEMOVERUSE# ---- --memoveruse ? -----------------------------------------------------
    # MEMOVERUSEif 'Python' in wisteria.globs.ARGS.memoveruse:
    # MEMOVERUSE    # pylint: disable=possibly-unused-variable
    # MEMOVERUSE    dumbstr = "0123456789"*100000
    # MEMOVERUSEif 'C++' in wisteria.globs.ARGS.memoveruse:
    # MEMOVERUSE    MemOverUse().memoveruse()

    # ---- main computation ---------------------------------------------------
    res = SerializationResult()

    if wisteria.globs.PLATFORM_SYSTEM == "Windows":
        mem0 = win_memory()
    else:
        mem0 = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss

    _error = False
    try:
        _res = module.encode(obj)
        _timeit = timeit.Timer('module.encode(obj)',
                               globals=locals())
        res.encoding_success = True
        res.encoding_strlen = _len(_res)
        res.encoding_time = _timeit.timeit(TIMEITNUMBER)
    except module.IaswnError as error:
        if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
            msgdebug(f"[{fingerprint}] '{module}': encoding failed ({error})")
        _error = True

    if not _error:
        try:
            _res2 = module.decode(_res)
            res.decoding_success = True
            _timeit = timeit.Timer("module.decode(_res)",
                                   globals=locals())
            res.decoding_time = _timeit.timeit(TIMEITNUMBER)

            if obj == _res2:
                res.reversibility = True
        except module.IaswnError as error:
            if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                msgdebug(f"[{fingerprint}] '{module}': decoding failed ({error})")
            res = None

    if res is not None and res.reversibility is True:
        if wisteria.globs.PLATFORM_SYSTEM == "Windows":
            res.mem_usage = win_memory() - mem0
        else:
            res.mem_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss - mem0

    return res


def serializer_json(action="serialize",
                    obj=None,
                    fingerprint=""):
    """
        serializer_json()

        Serializer for the json module.

        Like every serializer_xxx() function:
        * this function may return (action='version') the version of the concerned module.
        * this function may try (action='serialize') to encode/decode an <obj>ect.
        * if the serializer raises an error, this error is silently converted and no exception
          is raised. If an internal error happpens, a WisteriaError exception is raised.

        This function assumes that the concerned module has already be imported.

        _______________________________________________________________________

        ARGUMENTS:
        o  action: (str) either "version" either "serialize"
        o  obj:    the object to be serialized
        o  fingerprint: a string describing the operation (usefull to debug)

        RETURNED VALUE:
           - None if an error occcured
           - if <action> is (str)"version", return a string.
           - if <action> is (str)"serialize", return a SerializationResult object.
    """
    module = MODULES["json"]

    # -------------------
    # action == "version"
    # -------------------
    if action == "version":
        return module.__version__

    # ---------------------
    # action == "serialize"
    # ---------------------
    if action != "serialize":
        raise WisteriaError(f"(ERRORID031) Unknown 'action' keyword '{action}'.")

    # MEMOVERUSE# ---- --memoveruse ? -----------------------------------------------------
    # MEMOVERUSEif 'Python' in wisteria.globs.ARGS.memoveruse:
    # MEMOVERUSE    # pylint: disable=possibly-unused-variable
    # MEMOVERUSE    dumbstr = "0123456789"*100000
    # MEMOVERUSEif 'C++' in wisteria.globs.ARGS.memoveruse:
    # MEMOVERUSE    MemOverUse().memoveruse()

    # ---- main computation ---------------------------------------------------
    res = SerializationResult()

    if wisteria.globs.PLATFORM_SYSTEM == "Windows":
        mem0 = win_memory()
    else:
        mem0 = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss

    _error = False
    try:
        _res = module.dumps(obj)
        _timeit = timeit.Timer('module.dumps(obj)',
                               globals=locals())
        res.encoding_success = True
        res.encoding_strlen = _len(_res)
        res.encoding_time = _timeit.timeit(TIMEITNUMBER)
    except TypeError as error:
        if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
            msgdebug(f"[{fingerprint}] '{module}': encoding failed ({error})")
        _error = True

    if not _error:
        try:
            _res2 = module.loads(_res)
            res.decoding_success = True
            _timeit = timeit.Timer("module.loads(_res)",
                                   globals=locals())
            res.decoding_time = _timeit.timeit(TIMEITNUMBER)

            if obj == _res2:
                res.reversibility = True
        except (TypeError, AttributeError) as error:
            if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                msgdebug(f"[{fingerprint}] '{module}': decoding failed ({error})")
            res = None

    if res is not None and res.reversibility is True:
        if wisteria.globs.PLATFORM_SYSTEM == "Windows":
            res.mem_usage = win_memory() - mem0
        else:
            res.mem_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss - mem0

    return res


def serializer_jsonpickle(action="serialize",
                          obj=None,
                          fingerprint=""):
    """
        serializer_jsonpickle()

        Serializer for the jsonpickle module.

        Like every serializer_xxx() function:
        * this function may return (action='version') the version of the concerned module.
        * this function may try (action='serialize') to encode/decode an <obj>ect.
        * if the serializer raises an error, this error is silently converted and no exception
          is raised. If an internal error happpens, a WisteriaError exception is raised.

        This function assumes that the concerned module has already be imported.

        _______________________________________________________________________

        ARGUMENTS:
        o  action: (str) either "version" either "serialize"
        o  obj:    the object to be serialized
        o  fingerprint: a string describing the operation (usefull to debug)

        RETURNED VALUE:
           - None if an error occcured
           - if <action> is (str)"version", return a string.
           - if <action> is (str)"serialize", return a SerializationResult object.
    """
    module = MODULES["jsonpickle"]

    # -------------------
    # action == "version"
    # -------------------
    if action == "version":
        return module.__version__

    # ---------------------
    # action == "serialize"
    # ---------------------
    if action != "serialize":
        raise WisteriaError(f"(ERRORID032) Unknown 'action' keyword '{action}'.")

    # MEMOVERUSE# ---- --memoveruse ? -----------------------------------------------------
    # MEMOVERUSEif 'Python' in wisteria.globs.ARGS.memoveruse:
    # MEMOVERUSE    # pylint: disable=possibly-unused-variable
    # MEMOVERUSE    dumbstr = "0123456789"*100000
    # MEMOVERUSEif 'C++' in wisteria.globs.ARGS.memoveruse:
    # MEMOVERUSE    MemOverUse().memoveruse()

    # ---- main computation ---------------------------------------------------
    res = SerializationResult()

    if wisteria.globs.PLATFORM_SYSTEM == "Windows":
        mem0 = win_memory()
    else:
        mem0 = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss

    _error = False
    try:
        _res = module.dumps(obj)
        _timeit = timeit.Timer('module.dumps(obj)',
                               globals=locals())
        res.encoding_success = True
        res.encoding_strlen = _len(_res)
        res.encoding_time = _timeit.timeit(TIMEITNUMBER)
    except TypeError as error:
        if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
            msgdebug(f"[{fingerprint}] '{module}': encoding failed ({error})")
        _error = True

    if not _error:
        try:
            _res2 = module.loads(_res)
            res.decoding_success = True
            _timeit = timeit.Timer("module.loads(_res)",
                                   globals=locals())
            res.decoding_time = _timeit.timeit(TIMEITNUMBER)

            if obj == _res2:
                res.reversibility = True
        except (TypeError, AttributeError) as error:
            if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                msgdebug(f"[{fingerprint}] '{module}': decoding failed ({error})")
            res = None

    if res is not None and res.reversibility is True:
        if wisteria.globs.PLATFORM_SYSTEM == "Windows":
            res.mem_usage = win_memory() - mem0
        else:
            res.mem_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss - mem0

    return res


def serializer_jsonpickle_keystrue(action="serialize",
                                   obj=None,
                                   fingerprint=""):
    """
        serializer_jsonpickle_keystrue()

        Serializer for the jsonpickle module (keys=True).

        Like every serializer_xxx() function:
        * this function may return (action='version') the version of the concerned module.
        * this function may try (action='serialize') to encode/decode an <obj>ect.
        * if the serializer raises an error, this error is silently converted and no exception
          is raised. If an internal error happpens, a WisteriaError exception is raised.

        This function assumes that the concerned module has already be imported.

        _______________________________________________________________________

        ARGUMENTS:
        o  action: (str) either "version" either "serialize"
        o  obj:    the object to be serialized
        o  fingerprint: a string describing the operation (usefull to debug)

        RETURNED VALUE:
           - None if an error occcured
           - if <action> is (str)"version", return a string.
           - if <action> is (str)"serialize", return a SerializationResult object.
    """
    module = MODULES["jsonpickle"]

    # -------------------
    # action == "version"
    # -------------------
    if action == "version":
        return module.__version__

    # ---------------------
    # action == "serialize"
    # ---------------------
    if action != "serialize":
        raise WisteriaError(f"(ERRORID042) Unknown 'action' keyword '{action}'.")

    # MEMOVERUSE# ---- --memoveruse ? -----------------------------------------------------
    # MEMOVERUSEif 'Python' in wisteria.globs.ARGS.memoveruse:
    # MEMOVERUSE    # pylint: disable=possibly-unused-variable
    # MEMOVERUSE    dumbstr = "0123456789"*100000
    # MEMOVERUSEif 'C++' in wisteria.globs.ARGS.memoveruse:
    # MEMOVERUSE    MemOverUse().memoveruse()

    # ---- main computation ---------------------------------------------------
    res = SerializationResult()

    if wisteria.globs.PLATFORM_SYSTEM == "Windows":
        mem0 = win_memory()
    else:
        mem0 = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss

    _error = False
    try:
        _res = module.dumps(obj, keys=True)
        _timeit = timeit.Timer('module.dumps(obj, keys=True)',
                               globals=locals())
        res.encoding_success = True
        res.encoding_strlen = _len(_res)
        res.encoding_time = _timeit.timeit(TIMEITNUMBER)
    except TypeError as error:
        if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
            msgdebug(f"[{fingerprint}] '{module}': encoding failed ({error})")
        _error = True

    if not _error:
        try:
            _res2 = module.loads(_res, keys=True)
            res.decoding_success = True
            _timeit = timeit.Timer("module.loads(_res, keys=True)",
                                   globals=locals())
            res.decoding_time = _timeit.timeit(TIMEITNUMBER)

            if obj == _res2:
                res.reversibility = True
        except (TypeError, AttributeError) as error:
            if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                msgdebug(f"[{fingerprint}] '{module}': decoding failed ({error})")
            res = None

    if res is not None and res.reversibility is True:
        if wisteria.globs.PLATFORM_SYSTEM == "Windows":
            res.mem_usage = win_memory() - mem0
        else:
            res.mem_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss - mem0

    return res


def serializer_marshal(action="serialize",
                       obj=None,
                       fingerprint=""):
    """
        serializer_marshal()

        Serializer for the marshal module.

        Like every serializer_xxx() function:
        * this function may return (action='version') the version of the concerned module.
        * this function may try (action='serialize') to encode/decode an <obj>ect.
        * if the serializer raises an error, this error is silently converted and no exception
          is raised. If an internal error happpens, a WisteriaError exception is raised.

        This function assumes that the concerned module has already be imported.

        _______________________________________________________________________

        ARGUMENTS:
        o  action: (str) either "version" either "serialize"
        o  obj:    the object to be serialized
        o  fingerprint: a string describing the operation (usefull to debug)

        RETURNED VALUE:
           - None if an error occcured
           - if <action> is (str)"version", return a string.
           - if <action> is (str)"serialize", return a SerializationResult object.
    """
    module = MODULES["marshal"]

    # -------------------
    # action == "version"
    # -------------------
    if action == "version":
        return f"version {module.version}; (Python version) {sys.version.replace(chr(0x0A), '- ')}"

    # ---------------------
    # action == "serialize"
    # ---------------------
    if action != "serialize":
        raise WisteriaError(f"(ERRORID033) Unknown 'action' keyword '{action}'.")

    # MEMOVERUSE# ---- --memoveruse ? -----------------------------------------------------
    # MEMOVERUSEif 'Python' in wisteria.globs.ARGS.memoveruse:
    # MEMOVERUSE    # pylint: disable=possibly-unused-variable
    # MEMOVERUSE    dumbstr = "0123456789"*100000
    # MEMOVERUSEif 'C++' in wisteria.globs.ARGS.memoveruse:
    # MEMOVERUSE    MemOverUse().memoveruse()

    # ---- main computation ---------------------------------------------------
    res = SerializationResult()

    if wisteria.globs.PLATFORM_SYSTEM == "Windows":
        mem0 = win_memory()
    else:
        mem0 = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss

    _error = False
    try:
        _res = module.dumps(obj)
        _timeit = timeit.Timer('module.dumps(obj)',
                               globals=locals())
        res.encoding_success = True
        res.encoding_strlen = _len(_res)
        res.encoding_time = _timeit.timeit(TIMEITNUMBER)
    except (TypeError, ValueError) as error:
        if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
            msgdebug(f"[{fingerprint}] '{module}': encoding failed ({error})")
        _error = True

    if not _error:
        try:
            _res2 = module.loads(_res)
            res.decoding_success = True
            _timeit = timeit.Timer("module.loads(_res)",
                                   globals=locals())
            res.decoding_time = _timeit.timeit(TIMEITNUMBER)

            if obj == _res2:
                res.reversibility = True
        except (TypeError, AttributeError) as error:
            if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                msgdebug(f"[{fingerprint}] '{module}': decoding failed ({error})")
            res = None

    if res is not None and res.reversibility is True:
        if wisteria.globs.PLATFORM_SYSTEM == "Windows":
            res.mem_usage = win_memory() - mem0
        else:
            res.mem_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss - mem0

    return res


def serializer_pickle(action="serialize",
                      obj=None,
                      fingerprint=""):
    """
        serializer_pickle()

        Serializer for the pickle module.

        Like every serializer_xxx() function:
        * this function may return (action='version') the version of the concerned module.
        * this function may try (action='serialize') to encode/decode an <obj>ect.
        * if the serializer raises an error, this error is silently converted and no exception
          is raised. If an internal error happpens, a WisteriaError exception is raised.

        This function assumes that the concerned module has already be imported.

        _______________________________________________________________________

        ARGUMENTS:
        o  action: (str) either "version" either "serialize"
        o  obj:    the object to be serialized
        o  fingerprint: a string describing the operation (usefull to debug)

        RETURNED VALUE:
           - None if an error occcured
           - if <action> is (str)"version", return a string.
           - if <action> is (str)"serialize", return a SerializationResult object.
    """
    module = MODULES["pickle"]

    # -------------------
    # action == "version"
    # -------------------
    if action == "version":
        return f"(Python version) {sys.version.replace(chr(0x0A), '- ')}"

    # ---------------------
    # action == "serialize"
    # ---------------------
    if action != "serialize":
        raise WisteriaError(f"(ERRORID034) Unknown 'action' keyword '{action}'.")

    # MEMOVERUSE# ---- --memoveruse ? -----------------------------------------------------
    # MEMOVERUSEif 'Python' in wisteria.globs.ARGS.memoveruse:
    # MEMOVERUSE    # pylint: disable=possibly-unused-variable
    # MEMOVERUSE    dumbstr = "0123456789"*100000
    # MEMOVERUSEif 'C++' in wisteria.globs.ARGS.memoveruse:
    # MEMOVERUSE    MemOverUse().memoveruse()

    # ---- main computation ---------------------------------------------------
    res = SerializationResult()

    if wisteria.globs.PLATFORM_SYSTEM == "Windows":
        mem0 = win_memory()
    else:
        mem0 = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss

    _error = False
    try:
        _res = module.dumps(obj)
        _timeit = timeit.Timer('module.dumps(obj)',
                               globals=locals())
        res.encoding_success = True
        res.encoding_strlen = _len(_res)
        res.encoding_time = _timeit.timeit(TIMEITNUMBER)
    except TypeError as error:
        if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
            msgdebug(f"[{fingerprint}] '{module}': encoding failed ({error})")
        _error = True

    if not _error:
        try:
            _res2 = module.loads(_res)
            res.decoding_success = True
            _timeit = timeit.Timer("module.loads(_res)",
                                   globals=locals())
            res.decoding_time = _timeit.timeit(TIMEITNUMBER)

            if obj == _res2:
                res.reversibility = True
        except (TypeError, AttributeError) as error:
            if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                msgdebug(f"[{fingerprint}] '{module}': decoding failed ({error})")
            res = None

    if res is not None and res.reversibility is True:
        if wisteria.globs.PLATFORM_SYSTEM == "Windows":
            res.mem_usage = win_memory() - mem0
        else:
            res.mem_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss - mem0

    return res


def serializer_pyyaml(action="serialize",
                      obj=None,
                      fingerprint=""):
    """
        serializer_pyyaml()

        Serializer for the pyyaml module.

        Like every serializer_xxx() function:
        * this function may return (action='version') the version of the concerned module.
        * this function may try (action='serialize') to encode/decode an <obj>ect.
        * if the serializer raises an error, this error is silently converted and no exception
          is raised. If an internal error happpens, a WisteriaError exception is raised.

        This function assumes that the concerned module has already be imported.

        _______________________________________________________________________

        ARGUMENTS:
        o  action: (str) either "version" either "serialize"
        o  obj:    the object to be serialized
        o  fingerprint: a string describing the operation (usefull to debug)

        RETURNED VALUE:
           - None if an error occcured
           - if <action> is (str)"version", return a string.
           - if <action> is (str)"serialize", return a SerializationResult object.
    """
    module = MODULES["yaml"]

    # -------------------
    # action == "version"
    # -------------------
    if action == "version":
        return module.__version__

    # ---------------------
    # action == "serialize"
    # ---------------------
    if action != "serialize":
        raise WisteriaError(f"(ERRORID044) Unknown 'action' keyword '{action}'.")

    # MEMOVERUSE# ---- --memoveruse ? -----------------------------------------------------
    # MEMOVERUSEif 'Python' in wisteria.globs.ARGS.memoveruse:
    # MEMOVERUSE    # pylint: disable=possibly-unused-variable
    # MEMOVERUSE    dumbstr = "0123456789"*100000
    # MEMOVERUSEif 'C++' in wisteria.globs.ARGS.memoveruse:
    # MEMOVERUSE    MemOverUse().memoveruse()

    # ---- main computation ---------------------------------------------------
    res = SerializationResult()

    if wisteria.globs.PLATFORM_SYSTEM == "Windows":
        mem0 = win_memory()
    else:
        mem0 = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss

    _error = False
    try:
        _res = module.dump(obj, Dumper=module.Dumper)
        _timeit = timeit.Timer('module.dump(obj, Dumper=module.Dumper)',
                               globals=locals())
        res.encoding_success = True
        res.encoding_strlen = _len(_res)
        res.encoding_time = _timeit.timeit(TIMEITNUMBER)
    except (ValueError, TypeError) as error:
        if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
            msgdebug(f"[{fingerprint}] '{module}': encoding failed ({error})")
        _error = True

    if not _error:
        try:
            _res2 = module.load(_res, Loader=module.Loader)
            res.decoding_success = True
            _timeit = timeit.Timer("module.load(_res, Loader=module.Loader)",
                                   globals=locals())
            res.decoding_time = _timeit.timeit(TIMEITNUMBER)

            if obj == _res2:
                res.reversibility = True
        except (module.constructor.ConstructorError, ValueError) as error:
            if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                msgdebug(f"[{fingerprint}] '{module}': decoding failed ({error})")
            res = None

    if res is not None and res.reversibility is True:
        if wisteria.globs.PLATFORM_SYSTEM == "Windows":
            res.mem_usage = win_memory() - mem0
        else:
            res.mem_usage = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss - mem0

    return res


def init_serializers():
    """
        init_serializers()

        Initialize wisteria.globs.SERIALIZERS
    """
    for serializerdata in (
            SerializerData(
                name="iaswn",
                module_name="iaswn",
                human_name="Iaswn",
                internet="https://github.com/suizokukan/iaswn",
                func=serializer_iaswn),
            SerializerData(
                name="json",
                module_name="json",
                human_name="json",
                internet="https://docs.python.org/3/library/json.html",
                func=serializer_json),
            SerializerData(
                name="jsonpickle",
                module_name="jsonpickle",
                human_name="jsonpickle",
                internet="https://jsonpickle.github.io/",
                func=serializer_jsonpickle),
            SerializerData(
                name="jsonpickle_keystrue",
                module_name="jsonpickle",
                human_name="jsonpickle(keys=True)",
                internet="https://jsonpickle.github.io/",
                func=serializer_jsonpickle_keystrue),
            SerializerData(
                name="marshal",
                module_name="marshal",
                human_name="marshal",
                internet="https://docs.python.org/3/library/marshal.html#module-marshal",
                func=serializer_marshal),
            SerializerData(
                name="pickle",
                module_name="pickle",
                human_name="pickle",
                internet="https://docs.python.org/3/library/pickle.html",
                func=serializer_pickle),
            SerializerData(
                name="pyyaml",
                module_name="yaml",
                human_name="pyyaml",
                internet="https://pyyaml.org/",
                func=serializer_pyyaml),
            ):
        if trytoimport(serializerdata.module_name):
            wisteria.globs.SERIALIZERS[serializerdata.name] = serializerdata
            wisteria.globs.SERIALIZERS[serializerdata.name].version = serializerdata.func("version")
            if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                msgdebug(f"Successfully imported '{serializerdata.module_name}' module.")
        else:
            wisteria.globs.UNAVAILABLE_SERIALIZERS[serializerdata.name] = serializerdata
            if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                msgdebug(f"Could not import '{serializerdata.module_name}' module.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/results.py

    Fonctions to compute results, especially compute_results().

    ___________________________________________________________________________

    o  get_serializers_selection(serializer1, serializer2)
    o  get_data_selection(cmpdata, config)
    o  compute_results(config, serializer1, serializer2, cmpdata)
"""
from rich.console import Console
from rich.progress_bar import ProgressBar

import wisteria.data
import wisteria.globs
from wisteria.globs import VERBOSITY_MINIMAL, VERBOSITY_DETAILS, VERBOSITY_DEBUG
from wisteria.globs import PROGRESSBAR_LENGTH
from wisteria.wisteriaerror import WisteriaError
from wisteria.msg import msgdebug, msginfo, msgerror
from wisteria.serializers_classes import SerializationResults
from wisteria.utils import strdigest
from wisteria.reprfmt import fmt_serializer


def get_serializers_selection(serializer1,
                              serializer2):
    """
        get_serializers_selection()

        Return a tuple of all serializers defined by (str)<serializer1>, (str)<serializer2>.

        _______________________________________________________________________

        ARGUMENTS:
        o    <serializer1> : value returned by read_cmpstring()
        o    <serializer2> : value returned by read_cmpstring()

        RETURNED VALUE: a tuple of str, i.e. all serializers defined by (str)<serializer1>,
                        (str)<serializer2>
    """
    res = set()

    if serializer1 == "all":
        for serializer in wisteria.globs.SERIALIZERS:
            res.add(serializer)
    else:
        res.add(serializer1)

    if serializer2 == "all":
        for serializer in wisteria.globs.SERIALIZERS:
            res.add(serializer)
    else:
        res.add(serializer2)

    return tuple(res)


def get_data_selection(cmpdata,
                       config):
    """
        get_data_selection()

        Return a tuple of the data objects names selected by the <config>.

        _______________________________________________________________________

        ARGUMENTS:
        o  cmpdata: (str)"all" or "ini"
        o  config:  (dict) dict returned by read_cfgfile()

        RETURNED VALUE: (tuple) a tuple of (str)data keys
    """
    res = []

    if cmpdata == "all":
        res = tuple(wisteria.globs.DATA.keys())
    elif cmpdata == "ini":
        if config["data selection"]["data selection"] == "all":
            res = tuple(config["objects"].keys())
        elif config["data selection"]["data selection"] == "only if yes":
            res = (data_name for data_name in config["data objects"]
                   if config["data objects"][data_name])
        elif config["data selection"]["data selection"].startswith("data set/"):
            res = tuple(config["data sets"][config["data selection"]["data selection"]])
        else:
            raise WisteriaError(
                "(ERRORID020) "
                "Can't understand a value given to \\[data selection]'data selection': "
                f"what is '{config['data selection']['data selection']}' ? "
                "Known values are 'all', 'ini' and 'data set/xxx' "
                "where xxx is a string.")
    else:
        raise WisteriaError(
            "(ERRORID021) Can't understand a value given to the 'data' part in --cmp: "
            f"what is '{cmpdata}' ? "
            "Known values are 'all' and 'ini'. ")

    return res


def compute_results(config,
                    serializer1,
                    serializer2,
                    cmpdata):
    """
        compute_results()

        Create a SerializationResults object and try to fill it with all
        required encodings/decodings defined by <serializer1>,
        <serializer2> and <cmpdata>.

        _______________________________________________________________________

        ARGUMENTS:
        o  config               : None if no config file, otherwise the dict
                                  returned by read_cfgfile().
        o  (str)serializer1
        o  (str)serializer2
        o  (str)cmpdata         : 'all', 'ini' or 'cwc'

        RETURNED VALUE:    (SerializationResults, None) if no error occured
                        or (None, (int)exit_code) if an error occured
    """
    def erase_progress_bar():
        """
            erase_progress_bar()

            Erase the progress bar so that it is possible to write text over
            the ancient progress bar.

            Show the cursor.
        """
        if wisteria.globs.ARGS.verbosity not in (VERBOSITY_MINIMAL, VERBOSITY_DEBUG):
            # the following lines make the progress bar disappear.
            # the next rprint() will overwrite the spaces that are about
            # to be added:
            if PROGRESSBAR_LENGTH is None:
                console.file.write(" "*console.width)
            else:
                console.file.write(" "*PROGRESSBAR_LENGTH)
            console.file.write("\r")

            console.show_cursor(True)

    try:
        # serializers and data to be used through the tests:
        _serializers = get_serializers_selection(serializer1, serializer2)
        if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
            msgdebug(f"serializers to be used are: {_serializers}")
        _dataobjs = get_data_selection(cmpdata, config)
        if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
            msgdebug(f"data objs to be used are: {_dataobjs}")

        if wisteria.globs.ARGS.verbosity >= VERBOSITY_DETAILS:
            msginfo("Please wait until all required encodings/decodings have been computed.")

        results = SerializationResults()

        # (progress bar)
        # Please note that there can be NO progress bar if the debug mode is enabled:
        # the output can't display both correctly.
        # Please note that there can be NO progress bar if verbosity is set to minimal:
        # it's important for scripts calling this project from the outside.
        if wisteria.globs.ARGS.verbosity not in (VERBOSITY_MINIMAL, VERBOSITY_DEBUG):
            console = Console()
            progressbar = ProgressBar(width=PROGRESSBAR_LENGTH,
                                      total=len(_serializers)*len(_dataobjs))
            console.show_cursor(False)
            progressbar_index = 0

        for serializer in sorted(_serializers):
            results[serializer] = {}
            for data_name in sorted(_dataobjs):
                fingerprint = strdigest(serializer+data_name)

                if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                    msgdebug(f"About to call function for serializer='{serializer}' "
                             f"and data name='{data_name}' "
                             f"[{fingerprint}]")

                results[serializer][data_name] = wisteria.globs.SERIALIZERS[serializer].func(
                    action="serialize",
                    obj=wisteria.globs.DATA[data_name],
                    fingerprint=fingerprint)

                # (progress bar)
                # Please note that there can be NO progress bar if the debug mode is enabled:
                # the output can't display both correctly.
                # Please note that there can be NO progress bar if verbosity is set to minimal:
                # it's important for scripts calling this project from the outside.
                if wisteria.globs.ARGS.verbosity not in (VERBOSITY_MINIMAL, VERBOSITY_DEBUG):
                    progressbar_index += 1
                    progressbar.update(progressbar_index)
                    console.print(progressbar)
                    console.file.write("\r")

                if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
                    msgdebug(f"result: {results[serializer][data_name]} "
                             f"[{fingerprint}]")

            # we can't yet use results.total_encoding_strlen() since .finish_initialization()
            # has not be called:
            if sum(results[serializer][dataobj].encoding_strlen
                   for dataobj in results[serializer]
                   if results[serializer][dataobj] is not None and
                   results[serializer][dataobj].encoding_strlen is not None) == 0 and \
               not wisteria.globs.ARGS.tolerateabsurdvalues:
                erase_progress_bar()
                msgerror("(ERRORID043) Absurd value computed "
                         f"for serializer {fmt_serializer(serializer)} : "
                         "Σ sum(results[serializer][dataobj].encoding_strlen) is 0.")
                # (pimydoc)exit codes
                # ⋅*  0: normal exit code
                # ⋅*  1: normal exit code after --checkup
                # ⋅*  2: normal exit code after --downloadconfigfile
                # ⋅*  3: normal exit code after --mymachine
                # ⋅* -1: error, given config file can't be read (missing or ill-formed file)
                # ⋅* -2: error, ill-formed --cmp string
                # ⋅* -3: internal error, data can't be loaded
                # ⋅* -4: internal error, an error occured while computing the results
                # ⋅* -5: internal error, an error in main()
                # ⋅* -6: error, ill-formed --output string
                # ⋅* -7: error, an absurd value has been computed
                # ⋅* -8: error, missing required module
                # ⋅* -9: error, STR2REPORTSECTION_KEYS and STR2REPORTSECTION don't match
                return None, -7

        # (progress bar)
        # Please note that there can be NO progress bar if the debug mode is enabled:
        # the output can't display both correctly.
        # Please note that there can be NO progress bar if verbosity is set to minimal:
        # it's important for scripts calling this project from the outside.
        erase_progress_bar()

        if not results.finish_initialization():
            msgerror("(ERRORID015) Incorrect data, the program has to stop.")

            # (pimydoc)exit codes
            # ⋅*  0: normal exit code
            # ⋅*  1: normal exit code after --checkup
            # ⋅*  2: normal exit code after --downloadconfigfile
            # ⋅*  3: normal exit code after --mymachine
            # ⋅* -1: error, given config file can't be read (missing or ill-formed file)
            # ⋅* -2: error, ill-formed --cmp string
            # ⋅* -3: internal error, data can't be loaded
            # ⋅* -4: internal error, an error occured while computing the results
            # ⋅* -5: internal error, an error in main()
            # ⋅* -6: error, ill-formed --output string
            # ⋅* -7: error, an absurd value has been computed
            # ⋅* -8: error, missing required module
            # ⋅* -9: error, STR2REPORTSECTION_KEYS and STR2REPORTSECTION don't match
            return None, -3
        if results.dataobjs_number == 0:
            msginfo("No data to handle, the program can stop.")

            # (pimydoc)exit codes
            # ⋅*  0: normal exit code
            # ⋅*  1: normal exit code after --checkup
            # ⋅*  2: normal exit code after --downloadconfigfile
            # ⋅*  3: normal exit code after --mymachine
            # ⋅* -1: error, given config file can't be read (missing or ill-formed file)
            # ⋅* -2: error, ill-formed --cmp string
            # ⋅* -3: internal error, data can't be loaded
            # ⋅* -4: internal error, an error occured while computing the results
            # ⋅* -5: internal error, an error in main()
            # ⋅* -6: error, ill-formed --output string
            # ⋅* -7: error, an absurd value has been computed
            # ⋅* -8: error, missing required module
            # ⋅* -9: error, STR2REPORTSECTION_KEYS and STR2REPORTSECTION don't match
            return None, 2

        return results, None

    except WisteriaError as exception:
        msgerror(f"An error occured: {exception}")

        # (pimydoc)exit codes
        # ⋅*  0: normal exit code
        # ⋅*  1: normal exit code after --checkup
        # ⋅*  2: normal exit code after --downloadconfigfile
        # ⋅*  3: normal exit code after --mymachine
        # ⋅* -1: error, given config file can't be read (missing or ill-formed file)
        # ⋅* -2: error, ill-formed --cmp string
        # ⋅* -3: internal error, data can't be loaded
        # ⋅* -4: internal error, an error occured while computing the results
        # ⋅* -5: internal error, an error in main()
        # ⋅* -6: error, ill-formed --output string
        # ⋅* -7: error, an absurd value has been computed
        # ⋅* -8: error, missing required module
        # ⋅* -9: error, STR2REPORTSECTION_KEYS and STR2REPORTSECTION don't match
        return None, -4



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/globs.py

    Global definitions.

    ___________________________________________________________________________

    o  VERBOSITY_MINIMAL
    o  VERBOSITY_NORMAL
    o  VERBOSITY_DETAILS
    o  VERBOSITY_DEBUG

    o  ARGS

    o  TIMEITNUMBER

    o  TMPFILENAME

    o  REGEX_CMP
    o  REGEX_CMP__HELP

    o  REPORT_SHORTCUTS

    o  DEFAULT_CONFIG_FILENAME
    o  DEFAULTCFGFILE_URL

    o  MODULES

    o  SERIALIZERS

    o  DATA
    o  UNAVAILABLE_DATA

    o  FILECONSOLE_FILEOBJECT
    o  FILECONSOLE

    o  OUTPUT

    o  PROGRESSBAR_LENGTH

    o  UNITS

    o  LOGFILE_NAME

    o  STR2REPORTSECTION_KEYS
"""
import re


# values defined for --verbosity:
VERBOSITY_MINIMAL = 0
VERBOSITY_NORMAL = 1
VERBOSITY_DETAILS = 2
VERBOSITY_DEBUG = 3

# command line arguments
# will be set to argparse.ArgumentParser(...).parse_args()
ARGS = None

# number of times each serializer is called
TIMEITNUMBER = 1

# temp file default name
TMPFILENAME = "wisteria.tmp"

# regex used to parse the --cmp argument string
REGEX_CMP = re.compile(r"^\s*(?P<serializer1>[^\s\(\)]+)"
                       r"((\svs\s|\sversus\s|\sagainst\s)(?P<serializer2>[^\s\(\)]+))?"
                       r"(\s*\((?P<cmpdata>all|cwc|ini)\))?\s*$")
# string used by --help
REGEX_CMP__HELP = "all|serializer1[vs all|serializer2][(all|cwc|ini)]"

# REPORT_SHORTCUTS has two goals:
# (1) it translates a simple, human-readable string into a list of section parts
# (2) it gives all the accepted keywords that --report may understand.
REPORT_SHORTCUTS = {
    "bestof": "B1c;C2b;C2c;graphs;",
    "full": "titles;A;B;C;D1a;graphs;",
    "full+": "titles;A;B;C;D1b;graphs;",
    "full_debug": "titles;A;B;C;D;graphs;",
    "glance": "titles;B1b;C1a;C1b;C2a;C2b;C2c;graphs;",
    "laconic": "C;graphs;",
    "minimal": "C2c;graphs;",
    "none": "",
    }

# default name for the config file.
DEFAULT_CONFIG_FILENAME = "wisteria.ini"
# url of the default config file:
DEFAULTCFGFILE_URL = "https://raw.githubusercontent.com/suizokukan/wisteria/main/wisteria.ini"

# imported serializers modules
MODULES = {}

# dict storing all serializers used by the program.
#
# * format: SERIALIZERS[(str)serializer name] = SerializerData object
# * initialized by serializers.py::init_serializers()
SERIALIZERS = {}
# dict storing all serializers that the program can't use.
#
# * format: UNAVAILABLE_SERIALIZERS[(str)serializer name] = SerializerData object
# * initialized by serializers.py::init_serializers()
UNAVAILABLE_SERIALIZERS = {}

# dict storing all data used by the program.
#
# * format: DATA[(str)data name] = data object
# * initialized by data.py::init_data()
DATA = {}
# dict storing all data that the program can't use.
#
# * format: UNAVAILABLE_DATA[(str)data name] = string explaining why it's unavailable
# * initialized by data.py::init_data()
UNAVAILABLE_DATA = {}

# log file.
#

# Both variables are initialized by main.py()
FILECONSOLE_FILEOBJECT = None
#   value: rich.console.Console(file=...)
FILECONSOLE = None

# value of the --output argument
#  initialized by parse_output_argument(--output string)
# format:
#        ((bool)success, True if the --output string had been successfully parsed
#         (bool)output to the console ?,
#         (bool)output to the logfile ?,
#         (str)logfile open mode = 'a' or 'w',
#         (str)logfile name,
#        )
OUTPUT = []

# progress bar length (in characters)
# 'None' is an accepted value and extends the length of the bar
# to the entire width of the terminal.
PROGRESSBAR_LENGTH = None

# units used in this project
UNITS = {'time': 'seconds',
         'string length': 'characters',
         'memory': 'bytes',
         }

# name of the log file
LOGFILE_NAME = "report.txt"

# name of the graphs file
GRAPHS_FILENAME = "report__SUFFIX__.png"

#  globs.py:STR2REPORTSECTION_KEYS should be nothing but STR2REPORTSECTION.keys()
#
# Why those two variables ?
# Because the program needs to know STR2REPORTSECTION.keys() at step A
# before STR2REPORTSECTION has been initialized. Since this initialization
# requires modules that can't fit step A, we have to create a distinct list
# of keys.
#
# check_str2reportsection_keys() checks that keys from STR2REPORTSECTION and
# from STR2REPORTSECTION_KEYS are exactly the same.
STR2REPORTSECTION_KEYS = (
    'titles',
    'graphs',
    'A',
    'A1',
    'A2',
    'A3',
    'B',
    'B1',
    'B1a',
    'B1b',
    'B1c',
    'B1d',
    'B2',
    'B2a',
    'B2b',
    'C',
    'C1',
    'C1a',
    'C1b',
    'C2',
    'C2a',
    'C2b',
    'C2c',
    'D',
    'D1',
    'D1a',
    'D1b',
)


# = platform.system()
PLATFORM_SYSTEM = None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/utils.py

    BEWARE! Please use only verby basic imports since this file is used in
    "step A" section of the program, where no third-party libraries may
    be used.

    ___________________________________________________________________________

    o  get_missing_required_modules()
    o  normpath(path)
    o  shortenedstr(string, maximallength)
    o  trytoimport(module_name)
"""
# BEWARE! Please use only verby basic imports since this file is used in
# "step A" section of the program, where no third-party libraries may
# be used.
import hashlib
import importlib
import os
import os.path

import wisteria.globs


def get_missing_required_modules():
    """
        get_missing_required_modules()

        Return a list of the missing required modules.
        No debug message here since this function belongs to the 'A' step of the
        program.

        _______________________________________________________________________

        RETURNED VALUE: (list of str) a list of the missing required modules.
    """
    missing_modules = []
    for module_name in ('rich', 'psutil', 'cpuinfo'):
        if not trytoimport(module_name):
            missing_modules.append(module_name)
    return missing_modules


def normpath(path):
    """
        normpath()

        Return a human-readable (e.g. "~" -> "/home/myhome/") normalized
        version of a file path.

        ________________________________________________________________________

        PARAMETER : (str)path

        RETURNED VALUE : the normalized <path>
    """
    res = os.path.normpath(os.path.abspath(os.path.expanduser(path)))
    if res == ".":
        res = os.getcwd()
    return res


def shortenedstr(string,
                 maximallength=40):
    """
        This function reduces a <string> so that the maximum length of the returned
        string is equal to <maximallength>.

        ________________________________________________________________________

        PARAMETERS:
        o  (str)string       : to be string to be shortened
        o  (int)maximallength: the maximal length of <string>

        RETURNED VALUE: (str)<string> or a shortened version of <string>
    """
    # ---- nothing to do, <string> isn't too long. ----------------------------
    if len(string) <= maximallength:
        return string

    # ---- We have to reduce <string>. ----------------------------------------
    suffix = "[…]"

    # a special case: <maximallength> is really too short !
    if len(suffix) > maximallength:
        return string[:maximallength]

    # e.g. if maximallength==10, if string="01234567890"(len=11)  > "0123456[…]" (len=10)
    # e.g. if maximallength==10, if string="012345678901"(len=12) > "0123456[…]" (len=10)
    return string[:maximallength-len(suffix)] + suffix


def strdigest(string):
    """
        strdigest()

        Return the formatted fingerprint of <string>.

        _______________________________________________________________________

        ARGUMENT:
        o  (str)string: the string to be hashed

        RETURNED VALUE: (str)
    """
    return "0x"+hashlib.sha256(string.encode()).hexdigest()[:5]


def trytoimport(module_name):
    """
        trytoimport()

        Try to import <module_name> module.
        _______________________________________________________________________

        ARGUMENT:
        o  (str)module_name: the module to be imported

        RETURNED VALUE: (bool)success
    """
    res = True
    try:
        wisteria.globs.MODULES[module_name] = importlib.import_module(module_name)
    except ModuleNotFoundError:
        res = False
    return res



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/msg.py

    Fonctions used to print & log all projects' messages.

    msgreport() is reserved to checkup, report messages and mymachine().
    msgreporttitle() is reserved to title's report messages.

    Caveat!
    In this project, the choice of a msgxxx() method has nothing to do with
    priority (cf DEBUG, INFO, ERROR) but with message appearance (confer
    special symbols like '@', '>' or '!'). Priority is managed at each call
    of msgxxx() by a test of type
         if wisteria.globs.ARGS.verbosity == VERBOSITY_DEBUG:
             msgdebug(...)
    ___________________________________________________________________________

    o  _message(obj)
    o  msgcritical(obj="")
    o  msgdebug(obj="")
    o  msgerror(obj="")
    o  msginfo(obj="")
    o  msgreport(obj="")
    o  msgreporttitle(obj="")
    o  msgwarning(obj="")
"""
from rich import print as rprint

from wisteria.reprfmt import fmt_title, fmt_error
import wisteria.globs


def _message(obj):
    """
        _message()

        Internal function: displays and/or writes <obj> in the log file.

        _______________________________________________________________________

        ARGUMENT: <obj>, the object to be displayed/written in the log file
    """
    if wisteria.globs.OUTPUT[0]:
        rprint(obj)
    if wisteria.globs.OUTPUT[1]:
        wisteria.globs.FILECONSOLE.print(obj)


def msgcritical(obj=""):
    """
        msgcritical()

        Display and/or write <obj> in the log file; to be used like log.critical() .

        _______________________________________________________________________

        ARGUMENT: <obj>, the object to be displayed/written in the log file.
    """
    _message("!!! "+str(obj))


def msgdebug(obj=""):
    """
        msgcritical()

        Display and/or write <obj> in the log file; to be used like log.debug() .

        _______________________________________________________________________

        ARGUMENT: <obj>, the object to be displayed/written in the log file.
    """
    _message("@ "+str(obj))


def msgerror(obj=""):
    """
        msgerror()

        Display and/or write <obj> in the log file; to be used like log.error() .

        _______________________________________________________________________

        ARGUMENT: <obj>, the object to be displayed/written in the log file.
    """
    if isinstance(obj, str):
        obj = fmt_error(obj)
    _message(obj)


def msginfo(obj=""):
    """
        msginfo()

        Display and/or write <obj> in the log file; to be used like log.info() .

        _______________________________________________________________________

        ARGUMENT: <obj>, the object to be displayed/written in the log file.
    """
    _message("> "+str(obj))


def msgreport(obj=""):
    """
        msgreport()

        Display and/or write <obj> in the log file; to be used like log.info() .

        _______________________________________________________________________

        ARGUMENT: <obj>, the object to be displayed/written in the log file.
    """
    _message(obj)


def msgreporttitle(obj=""):
    """
        msgreporttitle()

        Display and/or write <obj> in the log file; to be used like log.info() .

        _______________________________________________________________________

        ARGUMENT: <obj>, the object to be displayed/written in the log file.
    """
    _message(fmt_title(obj))


def warning(obj=""):
    """
        msgwarning()

        Display and/or write <obj> in the log file; to be used like log.warning() .

        _______________________________________________________________________

        ARGUMENT: <obj>, the object to be displayed/written in the log file.
    """
    _message("! "+str(obj))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/serializers_classes.py

    All classes required to handle serializers.

    ___________________________________________________________________________

    o  SerializerData class
    o  SerializerDataObj class
    o  SerializationResult class
    o  SerializationResults class
"""
from dataclasses import dataclass

from wisteria.wisteriaerror import WisteriaError
from wisteria.reprfmt import fmt_serializer, fmt_ratio, fmt_time, fmt_nodata
from wisteria.reprfmt import fmt_stringlength, fmt_boolsuccess, fmt_mem_usage
from wisteria.msg import msgerror


class SerializerData:
    """
        SerializerData class

        SerializerData are used to store infos. about serializers in the SERIALIZERS dict.

        _______________________________________________________________________

        instance attributes:

        o  (str)name
        o  (str)human_name
        o  (str)internet
        o  (str)version
        o  (callable)func   : function to be called to use this serializer


        methods:

        o  __init__(self, name, module_name, human_name, internet, func)
        o  __repr__(self)
        o  checkup_repr(self)
        o  simple_repr(self)
    """
    def __init__(self,
                 name,
                 module_name,
                 human_name,
                 internet,
                 func):
        """
            SerializerData.__init__()

            ___________________________________________________________________

            ARGUMENTS:
            o  (str)name
            o  (str)module_name
            o  (str)human_name
            o  (str)internet
            o  (str)version
            o  (callable)func   : function to be called to use this serializer
        """
        self.name = name
        self.module_name = module_name
        self.human_name = human_name
        self.internet = internet
        self.version = None
        self.func = func

    def __repr__(self):
        """
            SerializerData.__repr__()

            For the check up output, see .checkup_repr().

            ___________________________________________________________________

            RETURNED VALUE: (str)a basic representation of <self>.
                            See also .checkup_repr() and .simple_repr()
        """
        return f"{self.name=}; {self.module_name=}; {self.human_name=}; {self.internet=}; " \
            f"{self.version=}; {self.func=}"

    def checkup_repr(self):
        """
            SerializerData.checkup_repr()

            Does the same as .__repr__() but for the check up output.

            ___________________________________________________________________

            RETURNED VALUE: (str)a basic representation of <self>.
                            See also .__repr__() and .simple_repr()
        """
        return f"{fmt_serializer(self.name)} ({self.version}), see {self.internet} ."

    def simple_repr(self):
        """
            SerializerData.simple_repr()

            Does the same as .__repr__() but for a simple output.

            ___________________________________________________________________

            RETURNED VALUE: (str)a basic representation of <self>.
                            See also .checkup_repr() and .__repr__()
        """
        return f"{fmt_serializer(self.name)} ({self.version})"


@dataclass
class SerializerDataObj:
    """
        SerializerDataObj class

        Nothing but an easy way to store (serializer, dataobj)

        _______________________________________________________________________

        methods:

        o  serializer: (str)
        o  dataobj:    (str)
    """
    serializer: str = None
    dataobj: str = None


# No useless public methods to add, indeed!
# pylint: disable=too-few-public-methods
class SerializationResult:
    """
        SerializationResult class

        Class used to store tests results in a SerializationResults object.

        _______________________________________________________________________

        instance attributes:
        o  (bool)  encoding_success
        o  (float) encoding_time
        o  (int)   encoding_strlen
        o  (bool)  decoding_success
        o  (float) decoding_time
        o  (bool)  reversibility
        o  (int)   mem_usage

        methods:

        o  __init__(self)
        o  __repr__(self)
    """
    def __init__(self):
        """
            SerializationResult.__init__()

            ___________________________________________________________________

            ARGUMENTS:

            o  (bool)  encoding_success
            o  (float) encoding_time
            o  (int)   encoding_strlen
            o  (bool)  decoding_success
            o  (float) decoding_time
            o  (bool)  reversibility
            o  (int)   mem_usage
        """
        self.encoding_success = False
        self.encoding_time = None
        self.encoding_strlen = None
        self.decoding_success = False
        self.decoding_time = None
        self.reversibility = False
        self.mem_usage = None

    def __repr__(self):
        """
            SerializationResult.__repr__()
        """
        return f"{self.encoding_success=}; {self.encoding_time=}; {self.encoding_strlen=}; " \
            f"{self.decoding_success=}; {self.decoding_time=}; {self.reversibility=}; " \
            "{self.mem_usage=}"


class SerializationResults(dict):
    """
        SerializationResults class

        SerializationResults objects are used to store all results created during the tests.

        A SerializationResults object has the following structure:
                SerializationResults[(str)serializer][(str)dataobj] = SerializationResult object

        Do not forget to call .finish_initialization() once you have finished initializing <self>.

        _______________________________________________________________________

        instance attributes:

        o  (list)serializers        : list of str
        o  (list)dataobjs           : list of str
        o  (int) serializers_number : =len(self.serializers)
        o  (int) dataobjs_number    : =len(self.dataobjs)
        o  (dict)hall         : e.g. {'encoding_success' = [(0.24, 'marshal'),
                                                                  (0.22, 'pickle')],
                                            ...

                                      keys are: 'encoding_success', 'encoding_time',
                                                'encoding_strlen',
                                                'decoding_success', 'decoding_time',
                                                'reversibility',
                                                'encoding_plus_decoding_time',
                                                'mem_usage'

        o  (dict)overallscores      : overallscores[serializer] = (int)overallscore


        methods:

        o  __init__(self)
        o  comparison_inside_hall(self, serializer, attribute)
        o  finish_initialization(self)
        o  get_base(self, attribute)
        o  get_dataobjs_base(self, attribute)
        o  get_hall(self, attribute, index)
        o  get_overallscore_rank(self, serializer)
        o  get_overallscore_bestrank(self)
        o  get_overallscore_worstrank(self)
        o  get_serializers_base(self, attribute)
        o  get_serializers_whose_overallscore_rank_is(self, rank)
        o  ratio_decoding_success(self, serializer=None, dataobj=None, output="fmtstr")
        o  ratio_encoding_success(self, serializer=None, dataobj=None, output="fmtstr")
        o  ratio_reversibility(self, serializer=None, dataobj=None, output="fmtstr")
        o  repr_attr(self, serializer, dataobj, attribute_name, output="fmtstr")
        o  total_decoding_time(self, serializer=None, dataobj=None, output="fmtstr")
        o  total_encoding_plus_decoding_time(self, serializer=None, dataobj=None, output="fmtstr")
        o  total_encoding_strlen(self, serializer=None, dataobj=None, output="fmtstr")
        o  total_encoding_time(self, serializer=None, dataobj=None, output="fmtstr")
        o  total_mem_usage(self, serializer=None, dataobj=None, output="fmtstr")
    """
    def __init__(self):
        """
            SerializationResults.__init__()

            ___________________________________________________________________

            ARGUMENTS:
            o  (list)serializers        : list of str
            o  (list)dataobjs           : list of str
            o  (int) serializers_number : =len(self.serializers)
            o  (int) dataobjs_number    : =len(self.dataobjs)
            o  (dict)hall         : e.g. {'encoding_success' = [(0.24, 'marshal'),
                                                                      (0.22, 'pickle')],
                                                ...

                                          keys are: 'encoding_success', 'encoding_time',
                                                    'encoding_strlen',
                                                    'decoding_success', 'decoding_time',
                                                    'reversibility',
                                                    'encoding_plus_decoding_time'
            o  (dict)overallscores      : overallscores[serializer] = (int)overallscore
        """
        dict.__init__(self)
        self.serializers = []
        self.dataobjs = []
        self.serializers_number = None
        self.dataobjs_number = None
        self.hall = None
        self.overallscores = None

    def comparison_inside_hall(self,
                               serializer,
                               attribute):
        """
            SerializationResults.comparison_inside_hall()

            Returns the serializers better placed than <serializer>
            (=placed before <serializer>) and the less well placed
            serializers than <serializer> (=placed after <serializer>)
            in relation to <attribute>.

            ___________________________________________________________________

            ARGUMENTS:
            o  (str)serializer
            o  (str)attribute

            RETURNED VALUE: (list of better placed serializers than <serializer>,
                             list of less well placed serializers than <serializer>)
        """
        _less = []
        _more = []

        before_serializer = False  # will be True when <serializer> will be read in the loop.
        for index in range(self.serializers_number):
            if self.hall[attribute][index][1] == serializer:
                before_serializer = True
            elif before_serializer:
                _less.append(self.hall[attribute][index][1])
            else:
                _more.append(self.hall[attribute][index][1])

        return _less, _more

    def finish_initialization(self):
        """
            SerializationResults.finish_initialization()

            Once the initialization of <self> is over, this method must be called to
            set self.self.serializers, self.dataobjs, self.serializers_number
            self.dataobjs_number, self.hall and self.overallscores

            ___________________________________________________________________

            RETURNED VALUE: (bool)success
        """
        self.serializers = sorted(self.keys())
        self.serializers_number = len(self.serializers)

        if self.serializers_number == 0:
            msgerror("(ERRORID016) Incorrect data, there's no serializer.")
            return False

        first_serializer = tuple(self.serializers)[0]
        self.dataobjs = sorted(self[first_serializer].keys())
        self.dataobjs_number = len(self.dataobjs)

        self.hall = {
            "encoding_success": sorted(((self.ratio_encoding_success(serializer=serializer,
                                                                     output="value"),
                                         serializer) for serializer in self.serializers),
                                       reverse=True),
            "encoding_time": sorted(((self.total_encoding_time(serializer=serializer,
                                                               output="value"),
                                      serializer) for serializer in self.serializers),
                                    reverse=False),
            "decoding_success": sorted(((self.ratio_decoding_success(serializer=serializer,
                                                                     output="value"),
                                         serializer) for serializer in self.serializers),
                                       reverse=True),
            "decoding_time": sorted(((self.total_decoding_time(serializer=serializer,
                                                               output="value"),
                                      serializer) for serializer in self.serializers),
                                    reverse=False),
            "encoding_strlen": sorted(((self.total_encoding_strlen(serializer=serializer,
                                                                   output="value"),
                                        serializer) for serializer in self.serializers),
                                      reverse=False),
            "reversibility": sorted(((self.ratio_reversibility(serializer=serializer,
                                                               output="value"),
                                      serializer) for serializer in self.serializers),
                                    reverse=True),
            "encoding_plus_decoding_time": sorted(((self.total_encoding_time(serializer=serializer,
                                                                             output="value") +
                                                    self.total_decoding_time(serializer=serializer,
                                                                             output="value"),
                                                    serializer) for serializer in self.serializers),
                                                  reverse=False),
            "mem_usage": sorted(((self.total_mem_usage(serializer=serializer,
                                                       output="value"),
                                  serializer) for serializer in self.serializers),
                                reverse=False),
            }

        # overall score computing:
        # * if a serializer is #0 (first rank) for a certain attribute,
        #   its score increases by results.serializers_number-0.
        # * if a serializer is #1 (first rank) for a certain attribute,
        #   its score increases by results.serializers_number-1.
        # * ...
        self.overallscores = {}
        for serializer in self.serializers:
            self.overallscores[serializer] = 0

            for attribute in ('encoding_plus_decoding_time',
                              'encoding_strlen',
                              'reversibility',
                              'mem_usage',
                              ):
                for index in range(self.serializers_number):
                    if self.hall[attribute][index][1] == serializer:
                        self.overallscores[serializer] += self.serializers_number-index

        return True

    def get_base(self,
                 attribute):
        """
            SerializationResults.get_base()

            Return <serializer, base data object> for attribute <attribute>.

            In other words, get_base() searches the first available couple
            of <serializer, base data object> which is initialized so that
            this couple could be a base for the <attribute>.

            ___________________________________________________________________

            ARGUMENT:
            o  (str)attribute, with only 4 values, namely "encoding_time",
               "decoding_time", "encoding_strlen" and "mem_usage"

            RETURNED VALUE: (SerializerDataObj), i.e. serializer + dataobj
        """
        assert attribute in ("encoding_time", "decoding_time", "encoding_strlen", "mem_usage")

        if attribute == "encoding_time":
            for serializer in self.serializers:
                for dataobj in self.dataobjs:
                    if self[serializer][dataobj] is not None and \
                       self[serializer][dataobj].encoding_time:
                        return SerializerDataObj(serializer=serializer, dataobj=dataobj)
            return SerializerDataObj()

        if attribute == "decoding_time":
            for serializer in self.serializers:
                for dataobj in self.dataobjs:
                    if self[serializer][dataobj] is not None and \
                       self[serializer][dataobj].decoding_time:
                        return SerializerDataObj(serializer=serializer, dataobj=dataobj)
            return SerializerDataObj()

        if attribute == "encoding_strlen":
            for serializer in self.serializers:
                for dataobj in self.dataobjs:
                    if self[serializer][dataobj] is not None and \
                       self[serializer][dataobj].encoding_strlen:
                        return SerializerDataObj(serializer=serializer, dataobj=dataobj)
            return SerializerDataObj()

        if attribute == "mem_usage":
            for serializer in self.serializers:
                for dataobj in self.dataobjs:
                    if self[serializer][dataobj] is not None and \
                       self[serializer][dataobj].mem_usage:
                        return SerializerDataObj(serializer=serializer, dataobj=dataobj)
            return SerializerDataObj()

        return None  # this line should never be executed.

    def get_hall(self,
                 attribute,
                 index):
        """
            SerializationResults.get_hall()

            Return the formatted string result of the 'hall of fame'
            for a given <attribute> and an <index>.
            In other words, answer the question: which serializer is number
            #<index> for a given <attribute> ?

            ___________________________________________________________________

            ARGUMENT:
            o  (str)attribute: 'encoding_success' or
                               'encoding_time' or
                               'decoding_success' or
                               'decoding_time' or
                               'encoding_strlen' or
                               'reversibility' or
                               'mem_usage' ?
            o  (int)index: 0 <= index < len(self.serializers_numbers-1)

            RETURNED VALUE: (str)a formatted string describing the result.
        """
        assert attribute in ('encoding_success',
                             'encoding_time',
                             'decoding_success',
                             'decoding_time',
                             'encoding_strlen',
                             'reversibility',
                             'mem_usage')

        serializer = self.hall[attribute][index][1]
        value = self.hall[attribute][index][0]

        if attribute == 'encoding_success':
            return f"{fmt_serializer(serializer)} " \
                f"[{self.ratio_encoding_success(serializer=serializer)}]"

        if attribute == 'encoding_time':
            return f"{fmt_serializer(serializer)} " \
                f"[{fmt_time(value)}]"

        if attribute == 'decoding_success':
            return f"{fmt_serializer(serializer)} " \
                f"[{self.ratio_decoding_success(serializer=serializer)}]"

        if attribute == 'decoding_time':
            serializer = self.hall[attribute][index][1]
            return f"{fmt_serializer(serializer)} " \
                f"[{fmt_time(value)}]"

        if attribute == 'reversibility':
            serializer = self.hall[attribute][index][1]
            return f"{fmt_serializer(serializer)} " \
                f"[{self.ratio_reversibility(serializer=serializer)}]"

        if attribute == 'encoding_strlen':
            serializer = self.hall[attribute][index][1]
            return f"{fmt_serializer(serializer)} " \
                f"[{fmt_stringlength(value)}]"

        if attribute == 'mem_usage':
            serializer = self.hall[attribute][index][1]
            return f"{fmt_serializer(serializer)} " \
                f"[{fmt_mem_usage(value)}]"

        return None  # this line should never be executed.

    def get_overallscore_rank(self,
                              serializer):
        """
            SerializationResults.get_overallscore_rank()

            Return the (int)rank of <serializer> among 'overall scores'.

            ___________________________________________________________________

            ARGUMENT:
            o  (str)serializer

            RETURNED VALUE: (int)rank, 0 <= rank < len(self.serializers_numbers)
        """
        _rank = None

        # for rank, (score, _serializer) ...
        for rank, (_, _serializer) in enumerate(sorted(
                ((self.overallscores[serializer],
                  serializer) for serializer in self.serializers), reverse=True)):
            if serializer == _serializer:
                _rank = rank

        return _rank

    def get_overallscore_bestrank(self):
        """
            SerializationResults.get_overallscore_bestrank()

            Return the list of the serializers with the best score
            among "overall scores". There's maybe more than one serializer
            having the highest score, hence the returned list.

            ___________________________________________________________________

            RETURNED VALUE: (list)list of (str)serializers
        """
        highestscore = sorted(
            ((self.overallscores[serializer],
              serializer) for serializer in self.serializers))[-1][0]
        res = []
        for serializer in self.serializers:
            if self.overallscores[serializer] == highestscore:
                res.append(serializer)
        return res

    def get_overallscore_worstrank(self):
        """
            SerializationResults.get_overallscore_worstrank()

            Return the list of the serializers with the worst score
            among "overall scores". There's maybe more than one serializer
            having the worst score, hence the returned list.

            ___________________________________________________________________

            RETURNED VALUE: (list)list of (str)serializers
        """
        worstscore = sorted(
            ((self.overallscores[serializer],
              serializer) for serializer in self.serializers))[0][0]
        res = []
        for serializer in self.serializers:
            if self.overallscores[serializer] == worstscore:
                res.append(serializer)
        return res

    def get_serializers_whose_overallscore_is(self,
                                              score):
        """
            SerializationResults.get_serializers_whose_overallscore_is()

            Return a list of all serializers whose overallscore is <score>.

            ___________________________________________________________________

            ARGUMENT: (int)score

            RETURNED VALUE: (list of str)list of serializers whose overallscore is <score>
        """
        return list(serializer for serializer in self.serializers
                    if self.overallscores[serializer] == score)

    def ratio_decoding_success(self,
                               serializer=None,
                               dataobj=None,
                               output="fmtstr"):
        """
            SerializationResults.ratio_decoding_success()

            Compute and format the ratio of decoding success for a <serializer>
            OR for a <dataobj>ect.

            _______________________________________________________________

            ARGUMENTS:
            o  <None|str>serializer: if not None, name of the serializer to be used.
            o  <None|str>dataobj: if not None, name of the data object to be used.
                BEWARE ! One and only one argument among <serializer> and <dataobj> can be set to
                         None.
            o  <output>(str): "fmtstr" for a formatted returned string or "value"
               for the (float)raw value

            RETURNED VALUE:
                (output=='fmtstr')a formatted string representing the input argument.
                (output=='value')a float
        """
        assert serializer is None or dataobj is None
        assert output in ('fmtstr', 'value')

        count = 0  # number of serializers or dataobjs that are taken in account.

        # serializer is not None:
        if serializer is not None:
            if self.serializers_number == 0:
                return fmt_ratio((None, None))

            for _dataobj in self[serializer]:
                if self[serializer][_dataobj] is not None and \
                   self[serializer][_dataobj].decoding_success:
                    count += 1
            if output == "fmtstr":
                return fmt_ratio((count, count/self.dataobjs_number))
            if output == "value":
                return count/self.dataobjs_number

        # dataobj is not None:
        if self.dataobjs_number == 0:
            return fmt_ratio((None, None))

        for _serializer in self:
            if self[_serializer][dataobj] is not None and \
               self[_serializer][dataobj].decoding_success:
                count += 1
        if output == "fmtstr":
            return fmt_ratio((count, count/self.serializers_number))
        if output == "value":
            return count/self.serializers_number

        return None  # this line should never be executed.

    def ratio_encoding_success(self,
                               serializer=None,
                               dataobj=None,
                               output="fmtstr"):
        """
            SerializationResults.ratio_encoding_success()

            Compute and format the ratio of encoding success for a <serializer>
            OR for a <dataobj>ect.

            _______________________________________________________________

            ARGUMENTS:
            o  <None|str>serializer: if not None, name of the serializer to be used.
            o  <None|str>dataobj: if not None, name of the data object to be used.
                BEWARE ! One and only one argument among <serializer> and <dataobj> can be set to
                         None.
            o  <output>(str): "fmtstr" for a formatted returned string or "value"
               for the (float)raw value

            RETURNED VALUE:
                (output=='fmtstr')a formatted string representing the input argument.
                (output=='value')a float
        """
        assert serializer is None or dataobj is None
        assert output in ("fmtstr", "value")

        count = 0  # number of serializers or dataobjs that are taken in account.

        # serializer is not None:
        if serializer is not None:
            if self.serializers_number == 0:
                return fmt_ratio((None, None))

            for _dataobj in self[serializer]:
                if self[serializer][_dataobj] is not None and \
                   self[serializer][_dataobj].encoding_success:
                    count += 1
            if output == "fmtstr":
                return fmt_ratio((count, count/self.dataobjs_number))
            if output == "value":
                return count/self.dataobjs_number

        # dataobj is not None:
        if self.dataobjs_number == 0:
            return fmt_ratio((None, None))

        for _serializer in self:
            if self[_serializer][dataobj] is not None and \
               self[_serializer][dataobj].encoding_success:
                count += 1
        if output == "fmtstr":
            return fmt_ratio((count, count/self.serializers_number))
        if output == "value":
            return count/self.serializers_number

        return None  # this line should never be executed.

    def ratio_reversibility(self,
                            serializer=None,
                            dataobj=None,
                            output="fmtstr"):
        """
            SerializationResults.ratio_reversibility()

            Compute and/or format the ratio of reversibility success for a <serializer>
            OR for a <dataobj>ect.

            _______________________________________________________________

            ARGUMENTS:
            o  <None|str>serializer: if not None, name of the serializer to be used.
            o  <None|str>dataobj: if not None, name of the data object to be used.
                BEWARE ! One and only one argument among <serializer> and <dataobj> can be set to
                         None.
            o  <output>(str): "fmtstr" for a formatted returned string or "value"
               for the (float)raw value

            RETURNED VALUE:
                (output=='fmtstr')a formatted string representing the input argument.
                (output=='value')a float
        """
        assert serializer is None or dataobj is None
        assert output in ('fmtstr', 'value')

        count = 0  # number of serializers or dataobjs that are taken in account.

        # serializer is not None:
        if serializer is not None:
            if self.serializers_number == 0:
                return fmt_ratio((None, None))

            for _dataobj in self[serializer]:
                if self[serializer][_dataobj] is not None and \
                   self[serializer][_dataobj].reversibility:
                    count += 1
            if output == "fmtstr":
                return fmt_ratio((count, count/self.dataobjs_number))
            if output == "value":
                return count/self.dataobjs_number

        # dataobj is not None:
        if self.dataobjs_number == 0:
            return fmt_ratio((None, None))

        for _serializer in self:
            if self[_serializer][dataobj] is not None and \
               self[_serializer][dataobj].reversibility:
                count += 1

        if output == "fmtstr":
            return fmt_ratio((count, count/self.serializers_number))
        if output == "value":
            return count/self.serializers_number

        return None  # this line should never be executed.

    def repr_attr(self,
                  serializer,
                  dataobj,
                  attribute_name,
                  output="fmtstr"):
        """
            SerializationResults.repr_attr()

            Format the value stored into self[serializer][dataobj].<attribute_name>.

            _______________________________________________________________

            ARGUMENTS:

            o  <str>serializer: name of the serializer to be used.
            o  <str>dataobj: name of the data object to be used.
            o  <str>attribute_name: 'decoding_success', 'decoding_time', 'encoding_strlen',
                                    'encoding_success', 'encoding_time',
                                    'reversibility', 'mem_usage'
            o  <output>(str): "fmtstr" for a formatted returned string

            RETURNED VALUE: a formatted string representing
                            self[serializer][dataobj].<attribute_name>
                (output=='fmtstr')a formatted string representing the input argument.
        """
        assert serializer is not None
        assert dataobj is not None
        assert attribute_name in ('decoding_success', 'decoding_time', 'encoding_strlen',
                                  'encoding_success', 'encoding_time',
                                  'reversibility', 'mem_usage')
        assert output in ('fmtstr',)

        res = None  # unexpected result !

        if attribute_name == "decoding_success":
            if output == "fmtstr":
                if self[serializer][dataobj] is None:
                    res = fmt_nodata()
                else:
                    res = fmt_boolsuccess(
                        self[serializer][dataobj].decoding_success)

        if attribute_name == "decoding_time":
            if output == "fmtstr":
                if self[serializer][dataobj] is None:
                    res = fmt_nodata()
                else:
                    res = fmt_time(
                        self[serializer][dataobj].decoding_time)

        if attribute_name == "encoding_strlen":
            if output == "fmtstr":
                if self[serializer][dataobj] is None or \
                   self[serializer][dataobj].encoding_strlen is None:
                    res = fmt_nodata()
                else:
                    res = fmt_stringlength(
                        self[serializer][dataobj].encoding_strlen)

        if attribute_name == "encoding_time":
            if output == "fmtstr":
                if self[serializer][dataobj] is None or \
                   self[serializer][dataobj].encoding_time is None:
                    res = fmt_nodata()
                else:
                    res = fmt_time(
                        self[serializer][dataobj].encoding_time)

        if attribute_name == "encoding_success":
            if output == "fmtstr":
                if self[serializer][dataobj] is None or \
                   self[serializer][dataobj].encoding_success is None:
                    res = fmt_nodata()
                else:
                    res = fmt_boolsuccess(
                        self[serializer][dataobj].encoding_success)

        if attribute_name == "reversibility":
            if output == "fmtstr":
                if self[serializer][dataobj] is None or \
                   self[serializer][dataobj].reversibility is None:
                    res = fmt_nodata()
                else:
                    res = fmt_boolsuccess(
                        self[serializer][dataobj].reversibility)

        if attribute_name == "mem_usage":
            if output == "fmtstr":
                if self[serializer][dataobj] is None or \
                   self[serializer][dataobj].mem_usage is None:
                    res = fmt_nodata()
                else:
                    res = fmt_mem_usage(
                        self[serializer][dataobj].mem_usage)

        return res

    def total_decoding_time(self,
                            serializer=None,
                            dataobj=None,
                            output="fmtstr"):
        """
            SerializationResults.total_decoding_time()

            Compute and format the total decoding time used by a <serializer>
            OR by a <dataobj>ect.

            _______________________________________________________________

            ARGUMENTS:
            o  <None|str>serializer: if not None, name of the serializer to be used.
            o  <None|str>dataobj: if not None, name of the data object to be used.
                BEWARE ! One and only one argument among <serializer> and <dataobj> can be set to
                         None.
            o  (str)output: output type and format
                    - "value": raw value (float)
                    - "fmtstr": formatted string (str)

            RETURNED VALUE: a formatted string representing the input arguments.
        """
        assert serializer is None or dataobj is None
        assert output in ('fmtstr', 'value',)

        res = None  # unexpected result !
        total = 0  # total time

        if serializer is not None:
            if self.serializers_number == 0:
                return fmt_time(None)

            for _dataobj in self[serializer]:
                if self[serializer][_dataobj] is not None and \
                   self[serializer][_dataobj].decoding_success:
                    total += self[serializer][_dataobj].decoding_time
            if output == "value":
                res = total
            elif output == "fmtstr":
                res = fmt_time(total)

        else:
            if self.dataobjs_number == 0:
                return fmt_time(None)

            for _serializer in self:
                if self[_serializer][dataobj] is not None and \
                   self[_serializer][dataobj].decoding_success:
                    total += self[_serializer][dataobj].decoding_time
            if output == "value":
                res = total
            elif output == "fmtstr":
                res = fmt_time(total)

        if res is None:
            raise WisteriaError("(ERRORID026) Internal error: the result could not be computed. "
                                f"{serializer=}; {dataobj=}; {output=};")
        return res

    def total_encoding_plus_decoding_time(self,
                                          serializer=None,
                                          dataobj=None,
                                          output="fmtstr"):
        """
            SerializationResults.total_encoding_plus_decoding_time()

            Compute and format the total encoding + decoding time used by a
            <serializer> OR by a <dataobj>ect.

            _______________________________________________________________

            ARGUMENTS:
            o  <None|str>serializer: if not None, name of the serializer to be used.
            o  <None|str>dataobj: if not None, name of the data object to be used.
                BEWARE ! One and only one argument among <serializer> and <dataobj> can be set to
                         None.
            o  (str)output: output type and format
                    - "value": raw value (float)
                    - "fmtstr": formatted string (str)

            RETURNED VALUE: a formatted string representing the input arguments.
        """
        assert serializer is None or dataobj is None
        assert output in ('fmtstr', 'value',)

        if output == "value":
            return self.total_encoding_time(serializer, dataobj, output) + \
                self.total_decoding_time(serializer, dataobj, output)
        if output == "fmtstr":
            return fmt_time(
                self.total_encoding_time(serializer, dataobj, output='value') +
                self.total_decoding_time(serializer, dataobj, output='value'))

        raise WisteriaError("(ERRORID027) Internal error: the result could not be computed. "
                            f"{serializer=}; {dataobj=}; {output=};")

    def total_encoding_strlen(self,
                              serializer=None,
                              dataobj=None,
                              output="fmtstr"):
        """
            SerializationResults.total_encoding_strlen()

            Compute and format the total encoding string length created by a <serializer>
            OR for a <dataobj>ect.

            _______________________________________________________________

            ARGUMENTS:
            o  <None|str>serializer: if not None, name of the serializer to be used.
            o  <None|str>dataobj: if not None, name of the data object to be used.
                BEWARE ! One and only one argument among <serializer> and <dataobj> can be set to
                         None.
            o  (str)output: output type and format
                    - "value": raw value (float)
                    - "fmtstr": formatted string (str)

            RETURNED VALUE: a formatted string representing the input arguments.
        """
        assert serializer is None or dataobj is None
        assert output in ('fmtstr', 'value',)

        res = None  # unexpected result !
        total = 0  # total time

        if serializer is not None:
            if self.serializers_number == 0:
                return fmt_stringlength(None)

            for _dataobj in self[serializer]:
                if self[serializer][_dataobj] is not None and \
                   self[serializer][_dataobj].encoding_strlen:
                    total += self[serializer][_dataobj].encoding_strlen
            if output == "value":
                res = total
            elif output == "fmtstr":
                res = fmt_stringlength(total)

        else:
            if self.dataobjs_number == 0:
                return fmt_stringlength(None)

            for _serializer in self:
                if self[_serializer][dataobj] is not None and \
                   self[_serializer][dataobj].encoding_strlen:
                    total += self[_serializer][dataobj].encoding_strlen
            if output == "value":
                res = total
            elif output == "fmtstr":
                res = fmt_stringlength(total)

        if res is None:
            raise WisteriaError("(ERRORID028) Internal error: the result could not be computed. "
                                f"{serializer=}; {dataobj=}; {output=};")
        return res

    def total_encoding_time(self,
                            serializer=None,
                            dataobj=None,
                            output="fmtstr"):
        """
            SerializationResults.total_encoding_time()

            Compute and format the total decoding time used by a <serializer>
            OR by a <dataobj>ect.

            _______________________________________________________________

            ARGUMENTS:
            o  <None|str>serializer: if not None, name of the serializer to be used.
            o  <None|str>dataobj: if not None, name of the data object to be used.
                BEWARE ! One and only one argument among <serializer> and <dataobj> can be set to
                         None.
            o  (str)output: output type and format
                    - "value": raw value (float)
                    - "fmtstr": formatted string (str)

            RETURNED VALUE: a formatted string representing the input argument.
        """
        assert serializer is None or dataobj is None
        assert output in ('fmtstr', 'value',)

        res = None  # unexpected result !
        total = 0  # total time

        if serializer is not None:
            if self.serializers_number == 0:
                return fmt_time(None)

            for _dataobj in self[serializer]:
                if self[serializer][_dataobj] is not None and \
                   self[serializer][_dataobj].encoding_success:
                    total += self[serializer][_dataobj].encoding_time
            if output == "value":
                res = total
            elif output == "fmtstr":
                res = fmt_time(total)

        else:
            if self.dataobjs_number == 0:
                res = fmt_time(None)

            for _serializer in self:
                if self[_serializer][dataobj] is not None and \
                   self[_serializer][dataobj].encoding_success:
                    total += self[_serializer][dataobj].encoding_time
            if output == "value":
                res = total
            elif output == "fmtstr":
                res = fmt_time(total)

        if res is None:
            raise WisteriaError("(ERRORID029) Internal error: the result could not be computed. "
                                f"{serializer=}; {dataobj=}; {output=};")
        return res

    def total_mem_usage(self,
                        serializer=None,
                        dataobj=None,
                        output="fmtstr"):
        """
            SerializationResults.total_mem_usage()

            Compute and format the total mem usage created by a <serializer>
            OR for a <dataobj>ect.

            _______________________________________________________________

            ARGUMENTS:
            o  <None|str>serializer: if not None, name of the serializer to be used.
            o  <None|str>dataobj: if not None, name of the data object to be used.
                BEWARE ! One and only one argument among <serializer> and <dataobj> can be set to
                         None.
            o  (str)output: output type and format
                    - "value": raw value (float)
                    - "fmtstr": formatted string (str)

            RETURNED VALUE: a formatted string representing the input arguments.
        """
        assert serializer is None or dataobj is None
        assert output in ('fmtstr', 'value',)

        res = None  # unexpected result !
        total = 0  # total time

        if serializer is not None:
            if self.serializers_number == 0:
                return fmt_mem_usage(None)

            for _dataobj in self[serializer]:
                if self[serializer][_dataobj] is not None and \
                   self[serializer][_dataobj].mem_usage:
                    total += self[serializer][_dataobj].mem_usage
            if output == "value":
                res = total
            elif output == "fmtstr":
                res = fmt_mem_usage(total)

        else:
            if self.dataobjs_number == 0:
                return fmt_mem_usage(None)

            for _serializer in self:
                if self[_serializer][dataobj] is not None and \
                   self[_serializer][dataobj].mem_usage:
                    total += self[_serializer][dataobj].mem_usage
            if output == "value":
                res = total
            elif output == "fmtstr":
                res = fmt_mem_usage(total)

        if res is None:
            raise WisteriaError("(ERRORID025) Internal error: the result could not be computed. "
                                f"{serializer=}; {dataobj=}; {output=};")
        return res



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
################################################################################
#    Wisteria Copyright (C) 2021 suizokukan
#    Contact: suizokukan _A.T._ orange dot fr
#
#    This file is part of Wisteria.
#    Wisteria is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Wisteria is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Wisteria.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    Wisteria project : wisteria/data.py

    All objects used as data by serializers are stored in the DATA dict.


    ___________________________________________________________________________

    * anyfunc()
    * init_data()
"""
import array
import calendar
import collections
import datetime
import decimal
import hashlib
import io
import numbers
import re
import time

import wisteria.globs
from wisteria.globs import TMPFILENAME
from wisteria.classesexamples.simpleclasses import MetaClass, RegularClass
from wisteria.classesexamples.simpleclasses import RegularClassInheritedDict
from wisteria.classesexamples.simpleclasses import RegularClassInheritedList


def anyfunc():
    """
        anyfunc()

        Fake function used by the DATA dict.
    """


def init_data():
    """
        init_data()

        Initialize wisteria.globs.DATA.

        PLEASE SYNCHRONIZE THIS LIST OF DATA WITH THE LIST IN THE
        CONFIG FILE(S) AND THE LIST IN README.MD.
    """
    wisteria.globs.DATA = {
        "array(b)":  array.array('b', (-1, 2)),
        "array(b/empty)": array.array('b'),
        "array(b_unsigned)": array.array('b', (1, 2)),
        "array(b_unsigned/empty)": array.array('B'),
        "array(u)": array.array('u', 'hello \u2641'),
        "array(u/empty)": array.array('u'),
        "array(h)": array.array('h', (-1, 2)),
        "array(h/empty)": array.array('h'),
        "array(h_unsigned)": array.array('H', (1, 2)),
        "array(h_unsigned/empty)": array.array('H'),
        "array(i)": array.array('i', (-1, 2)),
        "array(i/empty)": array.array('i'),
        "array(i_unsigned)": array.array('I', (1, 2)),
        "array(i_unsigned/empty)": array.array('I'),
        "array(l)": array.array('l', [-1, 2, 3, 4, 5]),
        "array(l/empty)": array.array('l'),
        "array(l_unsigned)": array.array('L', (1, 2)),
        "array(l_unsigned/empty)": array.array('L'),
        "array(q)": array.array('q', (-1, 2)),
        "array(q/empty)": array.array('q'),
        "array(q_unsigned)": array.array('Q', (1, 2)),
        "array(q_unsigned/empty)": array.array('Q'),
        "array(f)": array.array('f', (1.3, float('nan'))),
        "array(f/empty)": array.array('f'),
        "array(d)": array.array('d', [1.0, 2.0, 3.14]),
        "array(d/empty)": array.array('d'),

        "bool/false": False,
        "bool/true": True,

        "bytearray": bytearray(b"123"),
        "bytearray(empty)": bytearray(),

        "bytes": b"123",
        "bytes(empty)": b"",

        "calendar(calendar(3))": calendar.Calendar(3),

        "collections.chainmap(empty)": collections.ChainMap(),
        "collections.chainmap": collections.ChainMap({1: 2}, {2: 3}),
        "collections.counter(empty)": collections.Counter(),
        "collections.counter": collections.Counter((1, 2)),
        "collections.defaultdict(empty)": collections.defaultdict(),
        "collections.defaultdict": collections.defaultdict(None, {1: 2}),
        "collections.deque(empty)": collections.deque(),
        "collections.deque": collections.deque((1, 2)),
        "collections.ordereddict(empty)": collections.OrderedDict(),
        "collections.ordereddict": collections.OrderedDict({1: 2}),

        "complex": 1+2j,

        "datetime(datetime.datetime)": datetime.datetime(2001, 12, 1),
        "datetime(datetime.timedelta)":
        datetime.datetime(2001, 12, 1) - datetime.datetime(2000, 12, 1),

        "decimal(0.5)": decimal.Decimal(0.5),
        "decimal(1/7)": decimal.Decimal(1) / decimal.Decimal(7),
        "decimal(nan)": decimal.Decimal('NaN'),
        "decimal(-infinity)": decimal.Decimal("-Infinity"),
        "decimal(+infinity)": decimal.Decimal("+Infinity"),

        "dict(keys/bool)": {False: "False", True: "True"},
        "dict(keys/float)": {1.1: "value1.1", 2.2: "value2.2"},
        "dict(keys/int)": {0: "value0", 1: "value1", 2: "value2"},
        "dict(keys/str)": {"key1": "value1", "key2": "value2"},
        "dict(keys/str+subdicts)": {"key1": "value1", "key2": "value2", "key3": {"key4": "key4", }},

        "file descriptor": open(TMPFILENAME,      # pylint: disable=consider-using-with
                                encoding="utf-8"),

        "float": 1.1,
        "float(nan)": float('nan'),

        "frozenset": frozenset(("1", "2",)),
        "frozenset(empty)": frozenset(),

        "function": anyfunc,
        "function(python)": print,

        "hashlib(hashlib.sha1)": hashlib.sha1(b"some string"),
        "hashlib(hashlib.sha224)": hashlib.sha224(b"some string"),
        "hashlib(hashlib.sha256)": hashlib.sha256(b"some string"),
        "hashlib(hashlib.sha384)": hashlib.sha384(b"some string"),
        "hashlib(hashlib.sha512)": hashlib.sha512(b"some string"),

        "imported module": re,
        "imported module(class)": re.Pattern,
        "imported module(function)": re.sub,

        "int": 1,

        "io.string": io.StringIO(),
        "io.string(empty)": io.StringIO().write("string"),

        "list": ["1", "2", ],
        "list(empty)": [],
        "list(+sublists)": ["1", "2", ["3", ["4", ]]],

        "metaclass": MetaClass(),

        "memoryview": memoryview(b"123"),

        "none": None,

        "notimplemented": NotImplemented,

        "numbers(complex)": numbers.Complex,
        "numbers(integral)": numbers.Integral,
        "numbers(numbers)": numbers.Number(),
        "numbers(real)": numbers.Real,

        "pythonexception typeerror": TypeError,

        "range": range(1000),
        "range(empty)": range(0),

        "regularclass": RegularClass(),
        "regularclass(async_method)": RegularClass.async_method,
        "regularclass(class_method)": RegularClass.class_method,
        "regularclass(generator)": RegularClass.generator,
        "regularclass(method)": RegularClass.method,
        "regularclass(static_method)": RegularClass.static_method,
        "regularclassinheriteddict": RegularClassInheritedDict(),
        "regularclassinheritedlist": RegularClassInheritedList(),

        "re.match": re.match(".*", "abc"),
        "re.match(+flags)": re.match(".*", "abc", re.M),

        "re.pattern(bytes)": re.compile(".*"),
        "re.pattern(str)": re.compile(b".*"),

        "set": set(("1", "2",)),
        "set(empty)": set(),

        "str": "abc",
        "str(empty)": "",
        "str(long)": "abhg12234"*10000,
        "str(non ascii characters)": "êł¹@"+chr(0x1234)+chr(0x12345),

        "time(time.time)": time.time(),

        "tuple": ("1", "2",),
        "tuple(empty)": (),
        "tuple(+subtuples)": ("1", "2", ("3", ("4",))),

        "type(str)": str,
        "type(type(str))": type(str),
    }

    try:
        # The following 'import' statement has deliberately placed here and
        # not at the beginning of the file.
        # pylint: disable=import-outside-toplevel
        # pylint: disable=import-error
        import dateutil.parser
        wisteria.globs.DATA["dateutil(parser.parse)"] = dateutil.parser.parse("2021-03-04")
    except ImportError:
        wisteria.globs.UNAVAILABLE_DATA["dateutil(parser.parse)"] = "missing package: dateutil"
